#!/usr/bin/env ruby
if Process.uid == 628 #yhara
  $LOAD_PATH << '/home/yhara/temp/subversion-1.3.2/subversion/bindings/swig/ruby/'
else #www-data
  $LOAD_PATH << '/home/yhara/tmp/subversion-1.3.2/subversion/bindings/swig/ruby/'
end
# copied .libs/*.so to svn/ext/

require 'cgi'
require 'erb'
require 'kconv'
require 'svn/client'
require 'svn/repos'
require 'algorithm/diff'
require 'config.rb'

class Repotch
  include ERB::Util
  include Configuration
  CSS_URL = ""

  def initialize(name=nil)
    @repos_name = name
    @repos_uri  = REPOS_URI_OF[name] || REPOS_URI_OF.values.first
  end

  MIN = 60
  HOUR = 60*60
  DAY = 60*60*24
  MONTH = 60*60*24*30
  YEAR = 60*60*24*365
  def how_long(sec)
    case sec
    when 0...MIN
      "#{sec.to_i}秒"
    when MIN...HOUR
      "#{(sec/MIN).to_i}分"
    when HOUR...DAY
      "#{(sec/HOUR).to_i}時間"
    when DAY...(DAY*10)
      "#{'%0.1f' % (sec/DAY)}日"
    when (DAY*10)...MONTH
      "#{(sec/DAY).to_i}日"
    when MONTH...YEAR
      "#{'%.1f' % (sec/MONTH)}ヶ月"
    else
      if sec < 0
        "#{sec}秒"
      else
        "#{sec/YEAR}年#{(sec % YEAR)/MONTH}ヶ月"
      end
    end
  end
  
  def open_svn_session(repos_uri)
    ses = Svn::Ra::Session.open(repos_uri, {}, Svn::Ra::Callbacks.new(Svn::Client::Context.new.auth_baton))
    yield ses
  end

  class Commit
    include ERB::Util

    def initialize(*args)
      @repos_name,@paths,@rev,@author,@time,@how_old,@time_diff,@comment = *args
    end
    attr_accessor :paths, :rev, :author, :time, :how_old, :time_diff, :comment

    def to_html
      commit = self
      @paths = @paths.to_a.sort.reject{|path, info|
        (path == "/" || path == "/trunk") && info.action == "M"
      }
      ERB.new(File.read("template/_commit.rhtml")).result(binding)
    end
  end

  def get_logs(start_rev)
    @logs = []
    logs = []
    open_svn_session(@repos_uri) do |ses|
      @youngest_rev = ses.latest_revnum
      
      if start_rev.nil? || start_rev.to_i > @youngest_rev
        @start_rev = @youngest_rev
      else
        @start_rev = start_rev.to_i
      end

      n_revs = REVS_PER_PAGE + 1
      ses.log("", @start_rev, 0, n_revs) do |paths, rev, author, time, comment|
        logs << [paths, rev, author, time, comment]
      end
    end

    logs.size.times do |i|
      paths, rev, author, time, comment = *logs[i]
      com = Commit.new(@repos_name)
      com.paths     = paths
      com.rev       = rev
      com.author    = author
      com.time      = time
      com.how_old   = how_long(Time.now - time)
      com.time_diff = how_long(time - logs[i+1][3]) if logs[i+1]
      com.comment   = (comment=="" || comment.nil?) ? EMPTY_COMMENT_MES : comment.strip
      @logs << com
    end
    @logs.pop if @logs.size > REVS_PER_PAGE && @logs.last.rev != 0 #revision 0

    erb('template/logs.rhtml')
  end
  
  def diff(rev, path)
    open_svn_session(@repos_uri) do |ses|
      ses.log("", rev, 0, 1) do |paths, rev, author, time, comment|
        @commit = Commit.new(@repos_name, paths, rev, author, time, how_long(Time.now-time), nil, comment)
      end
    end

    if path
      @diffs = [ [path, get_diff(rev, path)] ]
    else
      @diffs = @commit.paths.select{|path, info| 
        info.action == "M" && (path != "/" && path != "/trunk")
      }.map{|path, info|
        [path, get_diff(rev, path)]
      }
    end
    
    erb('template/diff.rhtml')
  end

  def get_diff(rev, path)
    before = after = nil
    open_svn_session(@repos_uri) do |ses|
      after,  = *ses.file(path, rev)
      before, = *ses.file(path, rev-1)
    end

    #verb = []
    chunks = [[[]]]
    DiffRunner.sdiff(before.toutf8, after.toutf8) do |type,a,b|
      #verb << [type, a, b]
      if (chunks.last.last[0]==:matched) ^ (type==:matched) 
        chunks.push []
      end
      chunks.last.push [type, a, b]
    end
    #chunks.first.first.shift

    tmp = []
    chunks.each_with_index do |chunk, i|
      case chunk.last[0]
      when :added, :deleted, :changed
        tmp.push(chunks[i-1].last) if chunks[i-1]
        tmp.concat chunk
        tmp.push(chunks[i+1].last) if chunks[i+1]

        tmp.push [nil, nil,nil]
      end
    end

    buf = "<pre>"
    changed_after = ""
    tmp.each do |type, a, b|
      if type==:changed
        buf << "- #{h a}"
        changed_after << "+ #{h b}"
      else
        if changed_after != "" && type!=:deleted
          buf << changed_after
          changed_after = ""
        end
      end

      case type
      when nil
        buf << "<hr>\n"
      when :added
        buf << "+ #{h b}"
      when :deleted
        buf << "- #{h a}"
      when :matched
        buf << "  #{h a}"
      end
    end
    buf << "</pre>"
      #buf << "<table>" + verb.map{|type,a,b|
      #  type == :matched ? "" :
      # "<tr><td>#{type}</td><td>#{h a}</td><td>#{h b}</td>" 
      #}.join + "</table>"
  end

  class DiffRunner

    def self.sdiff(a,b)
      new(nil,nil,nil).sdiff(a.to_a, b.to_a) do |type, a, b|
        yield type, a, b
      end
    end

    def self.run(a,b)
      on_a = lambda{|i,j,v| yield :a,i,j,v}
      on_b = lambda{|i,j,v| yield :b,i,j,v}
      on_match = lambda{|i,j,v| yield :match,i,j,v}
      new(on_a, on_b, on_match).run(a.to_a, b.to_a)
    end
    
    def initialize(on_a, on_b, on_match = Proc.new)
      @on_a, @on_b, @on_match = on_a, on_b, on_match
    end

    def discard_a(i, j, v)         # a要素の処理
      @on_a.call(i,j,v)
    end
    def discard_b(i, j, v)
      @on_b.call(i,j,v)
    end
    def match(i, j, v)
      @on_match.call(i,j,v)
    end

    def run(a, b)
      d,snake,nn = Algorithm::Diff.ond(a, b) # 初期計算
      if d
        p = Algorithm::Diff.solve(a,b,d,snake,nn) # pathを計算
        Algorithm::Diff.traverse(a,b,path,self)         # トラバース実行
      end
    end

    def sdiff(a, b)
      d,snake,nn = Algorithm::Diff.ond(a, b) # 初期計算
      if d
        p = Algorithm::Diff.solve(a,b,d,snake,nn) # pathを計算
        Algorithm::Diff.sdiff(a,b,p).each {|x, a, b|
          yield x, a, b
        }
      end
    end

  end

  def list
    @repos_list = REPOS_URI_OF.map{|name, uri|
      commit = nil
      open_svn_session(uri) do |ses|
        ses.log("", ses.latest_revnum, 0, 1) do |paths, rev, author, time, comment|
          commit = Commit.new(name, paths, rev, author, time, how_long(Time.now-time), nil, comment)
        end
      end
      [name, commit]
    }.sort_by{|name, commit|
      - commit.time.to_i
    }
    erb("template/list.rhtml")
  end

  def erb(file, b = binding)
    ERB.new(File.read(file)).result(b)
  end

end

require 'pp'
cgi = CGI.new
cgi.out do 
  begin
    repos_name = cgi.params["repos"][0]
    start_rev = cgi.params["rev"][0] 
    diff_rev = cgi.params["diff"][0]
    path = cgi.params["path"][0]
    if diff_rev
      Repotch.new(repos_name).diff(diff_rev.to_i, path)
    elsif repos_name
      Repotch.new(repos_name).get_logs(start_rev)
    else
      Repotch.new.list
    end
  rescue Exception => e
    result = "<html><pre>[#{e.class}] #{CGI.escapeHTML e.message}<br>#{e.backtrace.join('<br>')}</pre></html>\n"
  end
end
